/*********************************************************************
 * Desktop Station air Sketch for DSair Hardware Platform
 *
 * Copyright (C) 2018 DesktopStation Co.,Ltd. / Yaasan
 * https://desktopstation.net/
 */

#define DEBUG

// include the library code:
#include <avr/wdt.h>
#include <EEPROM.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "Functions.h"
#include "FlashAirSharedMem.h"
#include "TrackReporterS88_DS.h"
#include "digitalWriteFast.h"

#define FIRMWARE_VER   '1' /* 0-9, a-z, A-Z*/

#define MAX_S88DECODER 1

#define RELPYERROR_300 "300 Command error"
#define RELPYERROR_301 "301 Syntax error"
#define RELPYERROR_302 "302 receive timeout"
#define RELPYERROR_303 "303 Unknown error"
#define RELPYERROR_304 "304 Serial error"
#define RELPYERROR_NONE ""


#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2
#define LEDSTATE_NOSD 3
#define LEDSTATE_ANALOG 4
#define PIN_PWMA 9
#define PIN_PWMB 10
#define PIN_EDC A1
#define PIN_RUNLED	2
#define PIN_ALERT 3
#define PIN_S88CHK A0
#define PIN_SPICS 4
#define PIN_CURRENT A2

#define SMEM_POWER		0
#define SMEM_ERROR		1
#define SMEM_EDC		2
#define SMEM_CUR		3
#define SMEM_SEQ		4

#define SHAREMEM_SIZE 48
#define SHAREMEM_STATUS_SIZE 264 //32 + 16 + 32x2 + 19 * 8= 112 + 144=264
#define SHAREMEM_ACCSIZE 32

#define THRESHOLD_EDC 276  // 1024d * (1.2 / 8) * 9V / 5V
#define THRESHOLD_EDC_DSA 375  // 1024d * (1.2 / 5.9) * 9V / 5V
#define THRESHOLD_OV  553  // 1024d * (1.2 / 8) * 20V / 5V
#define THRESHOLD_OV_DSA  999  // 1024d * (1.2 / 5.9) * 20V / 5V

#define HW_DSS1		0
#define HW_DSAIR1	1
#define HW_DSAIR2	2

#define MAX_LOCSTATUS_WEB	8

#ifdef __LGT8FX8P__
#define ADC_SHIFT 2
#else
#define ADC_SHIFT 0
#endif


typedef struct {
	byte mOV;
	byte mOC;
	byte mLV;
	byte mUnk1;
	byte mUnk2;
} DATA_STATUS;

typedef struct {
	word mAddress;//Marklin Address Format
	byte mSpeed;//(1/4 Data)
	byte mDir;
	word mFuncBuf1;
	word mFuncBuf2;
} DATA_LOC;


byte gHardware = HW_DSS1;
uint8_t buffer_sharedStatusMem[SHAREMEM_STATUS_SIZE];
uint8_t buffer_sharedMem[SHAREMEM_SIZE];
uint8_t buffer_lastcmd[SHAREMEM_SIZE];
uint8_t buffer_accessories[SHAREMEM_ACCSIZE];

/* 機関車データ確認 */
DATA_LOC gLocStatus[MAX_LOCSTATUS_WEB];

/* FlashAir接続確認 */
byte gFlag_FlashAir;

/* Decralation classes */
DSCoreLib DSCore;
TrackReporterS88_DS reporter(MAX_S88DECODER);


/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;
byte gLED_Counter = 0;

/* Sensor */
word gOffset_Current = 0;
word gEdc_data = 0;
short gCur_data = 0;
uint8_t gSwSensor = 0;
uint8_t gMaxS88Num = 0;

/* Error Threshold */
word gThreshold_OV = 0;
word gThreshold_LV = 0;
byte gLastErrorNo = 0;
uint8_t gAliveSeq = 0;

/* Parser */
String gRequest;
String function;
word arguments[4];
word numOfArguments;
word gLocAddresses[4] = {0,0,0,0};
word numOfLocAddr = 0;
boolean result;

DATA_STATUS gStatus;


//Task Schedule
unsigned long gPreviousL7 = 0; // 1000ms interval(Packet Task)
unsigned long gPreviousL5 = 0; //   50ms interval
unsigned long gPreviousL8 = 0; // 1000ms interval


void SetAnalogSpeed(word inSpeed, byte inDir);
void printString(const char *s, int x, int y);
void ReplyPowerPacket(byte inPower);
boolean SDIO_ReadSharedMem();
boolean SDIO_WriteSharedMem_Status();
boolean SDIO_WriteSharedMem_CVReply(word inCVNo, uint8_t inValue);
void SDIO_ClearSharedMem();

void LOCMNG_Clear(void);
byte LOCMNG_SetLocoSpeed(word inAddr, word inSpd);
byte LOCMNG_SetLocoDir(word inAddr, byte inDir);
byte LOCMNG_SetLocoFunc(word inAddr, byte inNo, byte inPower);
void LOCMNG_UpdateAcc(word inNo);

/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
	/* Debug message for PC */
	Serial.begin(115200);
	while (!Serial);
	
	Serial.println(F("Desktop Station air"));

	//初期化
	LOCMNG_Clear();
	
	//Pin mode
	pinMode(PIN_EDC, INPUT);//Edc
	pinMode(SENSOR_CURRENT, INPUT);
	pinModeFast(PIN_RUNLED, OUTPUT);
	pinModeFast(PIN_ALERT, INPUT);
	digitalWriteFast(PIN_ALERT, HIGH);
	
	pinModeFast(PIN_S88CHK, INPUT);
	digitalWriteFast(PIN_S88CHK, HIGH);
	
	/* DSCore初期化 */
	DSCore.Init();
	
	/* 線路に電源遮断 */
	DS_Power(0);
	
	/* Dammy sleep */
	CalcCurrentOffset();
	
	//Hardware seriesのチェック
	
	if( digitalReadFast(PIN_ALERT) == 0)
	{
		gHardware = HW_DSAIR2;
		gThreshold_OV = THRESHOLD_OV_DSA;
		gThreshold_LV = THRESHOLD_EDC_DSA;
		
	}
	else
	{
		if( digitalReadFast(PIN_S88CHK) == HIGH)
		{
			//DSair
			gThreshold_OV = THRESHOLD_OV_DSA;
			gThreshold_LV = THRESHOLD_EDC_DSA;
			gHardware = HW_DSAIR1;
		}
		else
		{
			//DSshield with FlashAir
			gThreshold_OV = THRESHOLD_OV;
			gThreshold_LV = THRESHOLD_EDC;
			gHardware = HW_DSS1;
		}
	}
	
	
	
	
	/* 共有メモリクリア */
	ClearShareMem();
	
	switch(gHardware)
	{
	case HW_DSS1:
		Serial.println(F("HW: DSshield with FlashAir"));
		break;
	case HW_DSAIR1:
		Serial.println(F("HW: DSair R1"));
		break;
	case HW_DSAIR2:
		Serial.println(F("HW: DSair R2"));
		break;
	}
	Serial.println(F("-------------"));
	
	// Initialize SD card.
	Serial.print(F("Init SD card..."));  
	
	if (SharedMemInit(PIN_SPICS) >= 0) 
	{
		Serial.println(F("OK"));
		gFlag_FlashAir = 1;
		
		//共有メモリをクリア
		SDIO_ClearSharedMem();
		SDIO_WriteSharedMem_Status();
		
	} else {
		
		Serial.println(F("Retry"));
		
		/* Retry */
		delay(500);
		
		if (SharedMemInit(PIN_SPICS) >= 0) 
		{
			Serial.println(F("OK"));
			gFlag_FlashAir = 1;
			
			//共有メモリをクリア
			SDIO_ClearSharedMem();
			SDIO_WriteSharedMem_Status();
			SDIO_WriteSharedMem_Acc();
		}
		else
		{
			/* 失敗 */
			Serial.println(F("NG"));
			gFlag_FlashAir = 0;
			gLED_State = LEDSTATE_NOSD;
		}
		
		
	}
	
	/* LED turn on */
	changeLED();
	
	Serial.println(F("100 Ready"));
	
	//Reset task
	gPreviousL5 = millis();
	gPreviousL8 = millis();
	gPreviousL7 = millis();
	
	//Timer1 490Hz -> 31.4kHz
	TCCR1B &= B11111000;
	TCCR1B |= B00000001;	
	
	
	sei();    //ENABLE INTERRUPTION
	
	/* ウォッチドッグ有効(1sec) */
	wdt_enable(WDTO_8S); // WDT reset

	
}


boolean parse(String inRequest) 
{
	int aPosSlash = -1;
	String request = inRequest;

	char aTemp[SHAREMEM_SIZE];
	request.toCharArray(aTemp, SHAREMEM_SIZE);

	int lpar = request.indexOf('(');
	if (lpar == -1) 
	{
		return false;
	}

	function = String(request.substring(0, lpar));
	function.trim();

	int offset = lpar + 1;
	int aAdddrOffset = 0;
	int comma = request.indexOf(',', offset);
	numOfArguments = 0;
	numOfLocAddr = 0;
	
	while (comma != -1) 
	{
		String tmp = request.substring(offset, comma);
		tmp.trim();
		
		/* スラッシュを探索する */
		aPosSlash = tmp.indexOf('/', 0);
		
		if( aPosSlash > 0)
		{
			tmp = tmp + "/";
			
			//Dammy値を指定
			arguments[numOfArguments++] = 0;
			aAdddrOffset = 0;
			
			//スラッシュはアドレス部だけに許可
			while( aPosSlash != -1)
			{
				String tmp2 = tmp.substring(aAdddrOffset, aPosSlash);
				tmp2.trim();
				
				if( numOfLocAddr < 4)
				{
					gLocAddresses[numOfLocAddr] = stringToWord(tmp2);
					numOfLocAddr++;
				}
				
				aAdddrOffset = aPosSlash + 1;
				aPosSlash = tmp.indexOf('/', aAdddrOffset);
			}
			
			offset = comma + 1;
			comma = request.indexOf(',', offset);
		}
		else
		{
			arguments[numOfArguments++] = stringToWord(tmp);
			offset = comma + 1;
			comma = request.indexOf(',', offset);
		}
	}

	int rpar = request.indexOf(')', offset);
	while (rpar == -1) 
	{
		return false;
	}

	if (rpar > offset) 
	{
		String tmp = request.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch() 
{
	byte aValue;
	boolean aResult;
	byte aDir = 0;
	
	if ((function == "setLocoDirection") || ((function == "DI")))
	{
		
		if( arguments[1] > 0)
		{
			//FWD=1,REV=2を FWD=0,REV=1に変換
			aDir = arguments[1] - 1;
			
			//異常値はREV固定
			if(aDir > 1)
			{
				aDir = 1;
			}
		}
		else
		{
			//FWD固定
			aDir = 0;
		}
		
		if( numOfLocAddr > 0)
		{
			//重連対応
			for( int i = 0; i < numOfLocAddr; i++)
			{
				LOCMNG_SetLocoDir(gLocAddresses[i], aDir);
				aResult =  DSCore.SetLocoDirection(gLocAddresses[i], aDir);
			}
			
			aResult = true;
		}
		else
		{
			//非重連
			LOCMNG_SetLocoDir(arguments[0], aDir);
			aResult =  DSCore.SetLocoDirection(arguments[0], aDir);
		}
		
		return aResult;
	}
	else if ((function == "setLocoFunction") || ((function == "FN")))
	{
		LOCMNG_SetLocoFunc(arguments[0], arguments[1], arguments[2]);
		
		return DSCore.SetLocoFunction(arguments[0], arguments[1] + 1, arguments[2]);
	
	}
	else if ((function == "setTurnout") || ((function == "TO")))
	{
		//テーブル更新
		//MM2: 1(0x3000) - 320(0x3140)
		//DCC: 1(0x3800) - 2044(0x3FFC)
		
		word aAccAddr = 0;
		
		if( (arguments[0] >= 0x3000) && (arguments[0] <= 0x3140))
		{
			aAccAddr = arguments[0] - 0x3000;
		}
		else if( (arguments[0] >= 0x3800) && (arguments[0] <= 0x3FFC))
		{
			aAccAddr = arguments[0] - 0x3800;
		}
		
		if( (aAccAddr >> 3) <= 31)
		{
			if( arguments[1] == 0)
			{
				buffer_accessories[aAccAddr >> 3] = buffer_accessories[aAccAddr >> 3] & ~(1UL << (aAccAddr & 0x07));
			}
			else
			{
				buffer_accessories[aAccAddr >> 3] = buffer_accessories[aAccAddr >> 3] | (1UL << (aAccAddr & 0x07));
			}
			
			LOCMNG_UpdateAcc(aAccAddr >> 3);
		}
		
		//ポイントパケット出力登録
		
		return DSCore.SetTurnout(arguments[0], (byte)arguments[1]);
	
	}
	else if ((function == "setPower") || ((function == "PW")))
	{
		
		if( arguments[0] == 1)
		{
			//端子を解放
			FreeAnalogSpeed();
			
			//電流オフセット計測
			CalcCurrentOffset();
			
			/* 状態変更 */
			gLED_State = LEDSTATE_RUN;
			
			//FlashAirのバッファにセット
			RegisterShareMem(SMEM_POWER, 1);
			RegisterShareMem(SMEM_ERROR, 1);
			
		}
		else
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_STOP;
			RegisterShareMem(SMEM_POWER, 0);
		}
		
		
		/* LED状態の変更 */
		gLED_Counter = 0;
		changeLED();
		
		return DSCore.SetPower(arguments[0]);
	
	}
	else if ((function == "setLocoSpeed") || ((function == "SP")))
	{
		if( numOfLocAddr > 0)
		{
			//重連対応
			for( int i = 0; i < numOfLocAddr; i++)
			{
				LOCMNG_SetLocoSpeed(gLocAddresses[i], arguments[1]);
				aResult =  DSCore.SetLocoSpeedEx(gLocAddresses[i], arguments[1], arguments[2]);
			}
			aResult = true;
		}
		else
		{
			//非重連
			LOCMNG_SetLocoSpeed(arguments[0], arguments[1]);
			aResult =  DSCore.SetLocoSpeedEx(arguments[0], arguments[1], arguments[2]);
		}
		
		return aResult;
	}
	else if (function == "DC")
	{
		//PRM0=Speed(0-1023), PRM1=Dir(0,1=FWD, 2=REV)
		if( (gLED_State == LEDSTATE_STOP) || (gLED_State == LEDSTATE_ANALOG))
		{
			
			if( arguments[0] > 0)
			{
				gLED_State = LEDSTATE_ANALOG;
			
				SetAnalogSpeed(arguments[0], arguments[1]);
			}
			else
			{
				SetAnalogSpeed(0, 0);
				gLED_State = LEDSTATE_STOP;
			}
		}
		
		return true;
	}
	else if ((function == "reset") || ((function == "RS")))
	{
		#ifdef DEBUG
		Serial.println(F("100 Ready"));
		#endif
		
		/* リセットパケットを投げる */
		if( gLED_State == LEDSTATE_RUN)
		{
			//内部のデータバッファを削除する
			DSCore.Clear();
			
			//リセット
			DSCore.SendReset();
			
		}
		
		return true;
	}
	/* reset */
	else if ((function == "setPing") || ((function == "PG")))
	{
		/* Type(2bytes), Version(2bytes), UID(4bytes) */
		Serial.println(F("@DSG,00,03,00,02,00,00,00,00"));
		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "GV")))
	{
		
		/* Clear Share Mem*/
		//SDIO_ClearSharedMem();
		
		//Clear last CV data
		SDIO_WriteSharedMem_CVReply(0, 0);

		/* LED turn on */
		digitalWrite(PIN_RUNLED, 1);
		
		/* Reset watchdog timer. */
		wdt_reset(); 

		aValue = 0;
		aResult = DSCore.ReadConfig( arguments[1], &aValue, 0, 0);
		
		
		//Send to Serial
		Serial.print("@CV,");
		Serial.print(arguments[0]);
		Serial.print(",");
		Serial.print(arguments[1]);
		Serial.print(",");
		
		if( aResult == true)
		{
			Serial.print(aValue);
			
			//Write to FlashAir
			SDIO_WriteSharedMem_CVReply(arguments[1], aValue);
			SDIO_WriteSharedMem_Status();
		}
		else
		{
			Serial.print(0xFFFF);
		}
		
		Serial.println(",");
		/* LED turn off */
		changeLED();
		
		
		return true;
	}
	else if ((function == "getS88")|| ((function == "gS8")))
	{
		int aMaxS88Num = MAX_S88DECODER;
		
		if( arguments[0] > 0)
		{
			aMaxS88Num = arguments[0];
		}
		
		gMaxS88Num = aMaxS88Num;
		
		return true;
		
	} /* getS88 */
	else if ((function == "setLocoConfig") || ((function == "SV")))
	{
		
		Serial.print("@CVWrite,");
		Serial.print(arguments[1]);
		Serial.print(",");
		Serial.print(arguments[2]);
		Serial.println("");
		
		/* LED turn on */
		digitalWrite(PIN_RUNLED, 1);
		
		DSCore.WriteConfig_Dir( arguments[1], arguments[2]);
		
		/* Clear Share Mem*/
		SDIO_ClearSharedMem();
		
	
		/* LED turn off */
		changeLED();

		return true;
	}
	else
	{
		return false;
	}
}

/***********************************************************************
* 
* Timer Interval
* 
************************************************************************/

void loop()
{
	long aEdc;
	signed long aCur;
	uint8_t aErrNo = 1;
	String aRecvCmd = "";
	char *aReplyText = RELPYERROR_NONE;
	static uint8_t sFlashAirAccess = 0;
	
	
	//パルスジェネレータのスキャン
	DSCore.Scan();
	
	if( (millis() - gPreviousL5) >= 200)
	{
		//Reset task
		gPreviousL5 = millis();
		
		//LED Control
		taskLED();
		
		/* Reset watchdog timer. */
		wdt_reset(); 
	}
	
	if( (millis() - gPreviousL8) >= 500)
	{
		
		//Reset task
		gPreviousL8 = millis();
		
		/* Sensor check */
		if( gSwSensor == 0)
		{
			gSwSensor = 1;
			gEdc_data = analogRead(PIN_EDC) >> ADC_SHIFT;
			
			/* OV check */
			if( gEdc_data > gThreshold_OV )
			{
				aErrNo = 4;
			}
			
			/* LV check */
			if( gEdc_data < gThreshold_LV )
			{
				aErrNo = 2;
				
			}
			
			// 1024d * (1.2 / 5.9) * Y[V] / 5[V] = X
			// -> 
			aEdc = (((long)gEdc_data * 246) >> 10);
			
			if( aEdc > 255)
			{
				aEdc = 255;
			}
			//Sensor Write
			RegisterShareMem(SMEM_EDC, aEdc);
			
		}
		else
		{
			gSwSensor = 0;
			gCur_data = analogRead(PIN_CURRENT) >> ADC_SHIFT;
			
			/* 電流の向きを正規化 */
			if( gCur_data >= gOffset_Current)
			{
				gCur_data = gCur_data - gOffset_Current;
			}
			else
			{
				gCur_data = gOffset_Current - gCur_data;
			}			
			
			aCur = ((long)gCur_data * 651) / 213;
			
			if( aCur > 1250)
			{
				aCur = 1250;
			}
			
			RegisterShareMem(SMEM_CUR, aCur);
			
		}
		
		
		/* DSairR2 HW  */
		if( (digitalReadFast(PIN_ALERT) == 1) && (gHardware == HW_DSAIR2))
		{
			//Over current at TB6642FG
			aErrNo = 8;
		}
		
		
		/* エラー処理 */
		if( aErrNo > 1)
		{
			if( DSCore.IsPower() == 1)
			{
				PowerOffByErr(aErrNo);
				//Serial.println(aErrNo);
			}
			
			//FlashAirのバッファにセット
			RegisterShareMem(SMEM_ERROR, aErrNo);
			
		}
		
		//送信回数
		RegisterShareMem(SMEM_SEQ, gAliveSeq);
		
		if( gAliveSeq >= 99)
		{
			gAliveSeq = 0;
		}
		else
		{
			gAliveSeq++;
		}
		
		//FlashAir 共有メモリSPI送信
		SDIO_WriteSharedMem_Status();
	}
	
	//パルスジェネレータのスキャン
	DSCore.Scan();
	
	
	if( sFlashAirAccess == 0)
	{
		
		/* コマンド受信＆処理 */
		boolean aRecvShareMem= SDIO_ReadSharedMem();
		
		if( aRecvShareMem == true)
		{
			//データをシリアルに流し込む
			
			for( int i = 0; i < SHAREMEM_SIZE; i++)
			{
				if( buffer_sharedMem[i] != ')' )
				{
					char aWord = (char)(buffer_sharedMem[i]);
					
					aRecvCmd = aRecvCmd + aWord;
				}
				else
				{
					aRecvCmd = aRecvCmd + ")";
					break;
				}
			}
			
			if (parse(aRecvCmd)) 
			{
				#ifdef DEBUG
				Serial.println(aRecvCmd);
				#endif
				
				if (dispatch()) 
				{
					aReplyText = "200 Ok";
				}
				else 
				{
					aReplyText = RELPYERROR_300;
				}
			} 
			else 
			{
				aReplyText = RELPYERROR_301;
			}
			
			#ifdef DEBUG
			Serial.println(aReplyText);
			#endif
		}
		else
		{
			/* Nothing to do */
			
			
		}
	}
	
	sFlashAirAccess = sFlashAirAccess + 1;
	
	if( sFlashAirAccess > 3)
	{
		sFlashAirAccess = 0;
	}
	
}

void ScanS88(void)
{

	if( gMaxS88Num > 0)
	{
		reporter.refresh(gMaxS88Num);

		word aFlags = 0;

		for( int j = 0; j < gMaxS88Num; j++)
		{
		  aFlags = (reporter.getByte((j << 1) + 1) << 8) + reporter.getByte(j << 1);
		}
	}
}



void PowerOffByErr(uint8_t inErrNo)
{
	DSCore.SetPower(0);
	delay(10);
	ReplyPowerPacket(0);
	DSCore.SetPower(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
	gLED_Counter =0;
	gLastErrorNo = inErrNo;
	
	/* FlashAirにも反映 */
	RegisterShareMem(SMEM_POWER, 0);
}

void ReplyPowerPacket(byte inPower)
{
	#ifdef DEBUG
	/* Power */
	Serial.print("@PWR,");
	Serial.print(inPower);
	Serial.println(",");
	#endif
}


void changeLED(void)
{
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		digitalWriteFast(PIN_RUNLED, 0);
		break;
		
	case LEDSTATE_RUN:
		digitalWriteFast(PIN_RUNLED, 1);
		break;
	
	case LEDSTATE_ANALOG:
		break;
		
	case LEDSTATE_ERR:
		break;
		
	case LEDSTATE_NOSD:
		break;
	}
}

void taskLED(void)
{
	
	/* カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	gLED_Counter = gLED_Counter & 0b1111;
	
	//LED操作
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		
		if( gLED_Counter <= 3)
		{
			digitalWriteFast(PIN_RUNLED, 1);
		}
		else
		{
			digitalWriteFast(PIN_RUNLED, 0);
		}
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		
		if( gLED_Toggle == 0)
		{
			digitalWriteFast(PIN_RUNLED, 0);
		}
		else
		{
			digitalWriteFast(PIN_RUNLED, 1);
		}
		
		gLED_Toggle = (~gLED_Toggle) & 1;
		break;
	
	case LEDSTATE_ANALOG:
		if( (gLED_Counter == 0) || (gLED_Counter == 8))
		{
			digitalWriteFast(PIN_RUNLED, 0);
		}
		else
		{
			digitalWriteFast(PIN_RUNLED, 1);
		}
		
		break;
		
	case LEDSTATE_NOSD:
		
		if( gLED_Counter <= 0)
		{
			digitalWriteFast(PIN_RUNLED, 1);
		}
		else if(gLED_Counter == 2)
		{
			digitalWriteFast(PIN_RUNLED, 1);
		}
		else
		{
			digitalWriteFast(PIN_RUNLED, 0);
		}
		
		break;
	}
}

void SetAnalogSpeed(word inSpeed, byte inDir)
{
	
	word aSpeed = inSpeed >> 2;
	
	if( aSpeed > 255)
	{
		aSpeed = 255;
	}
	
	if( inSpeed == 0)
	{
		FreeAnalogSpeed();
	}
	else
	{
		if( inDir == 2)
		{
			//REV
			analogWrite(PIN_PWMA, 255 - aSpeed);
			analogWrite(PIN_PWMB, 255);
		}
		else
		{
			//FWD
			analogWrite(PIN_PWMB, 255 - aSpeed);
			analogWrite(PIN_PWMA, 255);
		}
	}
}

void FreeAnalogSpeed()
{
	analogWrite(PIN_PWMB, 0);
	analogWrite(PIN_PWMA, 0);
}

void SDIO_ClearSharedMem()
{
	
	//Clear Buffer
	memset(buffer_sharedMem, 0, SHAREMEM_SIZE);
	
	if( SharedMemWrite( 0, SHAREMEM_SIZE, buffer_sharedMem))
	{
		Serial.println(F("\nwrite err"));
		return;
	}	
	
	//Update LastCommandBuffer
	for( byte i = 0; i < SHAREMEM_SIZE; i++)
	{
		buffer_lastcmd[i] = 0;
	}
	
}

boolean SDIO_ReadSharedMem()
{
	boolean aResultFnc = false;
	uint8_t aChk = 0;
	
	if( gFlag_FlashAir == 0)
	{
		return aResultFnc;
	}
	
	//Clear Buffer
	memset(buffer_sharedMem, 0, SHAREMEM_SIZE);
	
	if( SharedMemRead(0, SHAREMEM_SIZE, buffer_sharedMem))
	{
		Serial.println("\nread err");
		return false;
	}
	
	//Check between received command and last one.
	for( byte i = 0; i < SHAREMEM_SIZE; i++)
	{
		if( buffer_lastcmd[i] != buffer_sharedMem[i])
		{
			aChk++;
		}
	}
	
	if( buffer_sharedMem[0] == 0x00)
	{
		//Shared memory is empty
		aResultFnc = false;
	}
	else if( aChk == 0)
	{
		//Same command received.
		aResultFnc = false;
	}
	else
	{
		aResultFnc = true;
		
		//Update LastCommandBuffer
		for( byte i = 0; i < SHAREMEM_SIZE; i++)
		{
			buffer_lastcmd[i] = buffer_sharedMem[i];
		}
		
	}
	
	return aResultFnc;
}


boolean SDIO_WriteSharedMem_Status()
{
	
	if( gFlag_FlashAir == 0)
	{
		return false;
	}
	
	if( SharedMemWrite( 0x80, SHAREMEM_STATUS_SIZE, buffer_sharedStatusMem))
	{
		Serial.println(F("\nwrite err"));
		
		return false;
	}
	else
	{
		return true;
	}
}

void ClearShareMem()
{
	
	buffer_sharedStatusMem[0] = 0x4e;//PWR
	buffer_sharedStatusMem[1] = 0x2c;//カンマ
	buffer_sharedStatusMem[2] = 0x30;//ERR
	buffer_sharedStatusMem[3] = 0x2c;//カンマ
	buffer_sharedStatusMem[4] = FIRMWARE_VER;//VER
	buffer_sharedStatusMem[5] = 0x2c;//カンマ
	buffer_sharedStatusMem[6] = 0x30;//制御車両数(0-15, 16進数)
	buffer_sharedStatusMem[7] = 0x2c;//カンマ
	buffer_sharedStatusMem[8] = 0x30;//電圧
	buffer_sharedStatusMem[9] = 0x30;
	buffer_sharedStatusMem[10] = 0x30;
	buffer_sharedStatusMem[11] = 0x2c;//カンマ
	buffer_sharedStatusMem[12] = 0x30;//電流
	buffer_sharedStatusMem[13] = 0x30;
	buffer_sharedStatusMem[14] = 0x2c;//カンマ
	buffer_sharedStatusMem[15] = 0x30 + gHardware;//HW version
	buffer_sharedStatusMem[16] = 0x2c;//カンマ
	buffer_sharedStatusMem[17] = 0x30;//送信回数
	buffer_sharedStatusMem[18] = 0x30;//送信回数（生存チェック)
	buffer_sharedStatusMem[19] = 0x2c;//カンマ
	buffer_sharedStatusMem[20] = 0x30;
	buffer_sharedStatusMem[21] = 0x30;
	buffer_sharedStatusMem[22] = 0x30;
	buffer_sharedStatusMem[23] = 0x30;
	buffer_sharedStatusMem[24] = 0x30;
	buffer_sharedStatusMem[25] = 0x30;
	buffer_sharedStatusMem[26] = 0x30;
	buffer_sharedStatusMem[27] = 0x30;
	buffer_sharedStatusMem[28] = 0x30;
	buffer_sharedStatusMem[29] = 0x30;
	buffer_sharedStatusMem[30] = 0x30;
	buffer_sharedStatusMem[31] = 0x3b;
	
	//CV
	for(int i = 0; i < 15; i++)
	{
		buffer_sharedStatusMem[32 + i] = 0x30;
	}
	
	buffer_sharedStatusMem[47] = 0x3b;
	
	
	//ポイント
	
	for(int i = 0; i < SHAREMEM_ACCSIZE; i++)
	{
		buffer_sharedStatusMem[48 + i * 2] = 0x30;
		buffer_sharedStatusMem[48 + i * 2 + 1] = 0x30;
	}
	
		buffer_sharedStatusMem[112] = 0x3b;	//;
	
	
	//アクティブな車両データ(最新5つ)
	
	for(int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		buffer_sharedStatusMem[113 + i * 19] = 0x30;	//LocAddr
		buffer_sharedStatusMem[113 + i * 19 + 1] = 0x30;//LocAddr
		buffer_sharedStatusMem[113 + i * 19 + 2] = 0x30;//LocAddr
		buffer_sharedStatusMem[113 + i * 19 + 3] = 0x30;//LocAddr
		buffer_sharedStatusMem[113 + i * 19 + 4] = 0x2C;//,
		buffer_sharedStatusMem[113 + i * 19 + 5] = 0x30;//Spd
		buffer_sharedStatusMem[113 + i * 19 + 6] = 0x30;//Spd
		buffer_sharedStatusMem[113 + i * 19 + 7] = 0x2C;//,
		buffer_sharedStatusMem[113 + i * 19 + 8] = 0x30;//Dir
		buffer_sharedStatusMem[113 + i * 19 + 9] = 0x2C;//,
		buffer_sharedStatusMem[113 + i * 19 + 10] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 11] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 12] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 13] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 14] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 15] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 16] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 17] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 18] = 0x2f;///
	}	
	
	
}

void RegisterShareMem(uint8_t inNumber, uint8_t inValue)
{
	/*
	0x1000-0x107F FlashAir→Arduino コマンド
	0x1080-0x10AF ステータス
	0x10B0-0x10C0 CV Reply
	*/
	
	char aTxtBuf[5];
	
	if( gFlag_FlashAir == 0)
	{
		return;
	}
	
	
	switch(inNumber)
	{
	case SMEM_POWER:
		//Power
		buffer_sharedStatusMem[0] = (inValue == 1) ? 0x59 : 0x4e;
		break;
		
	case SMEM_ERROR:
		//ErrorCode
		sprintf(aTxtBuf,"%d",inValue);//0x00は無効
		buffer_sharedStatusMem[2] = aTxtBuf[0];
		break;
		
	case SMEM_EDC:
		//電圧
		sprintf(aTxtBuf,"%03d",inValue);//0x00は無効
		buffer_sharedStatusMem[8] = aTxtBuf[0];
		buffer_sharedStatusMem[9] = aTxtBuf[1];
		buffer_sharedStatusMem[10] = aTxtBuf[2];
		
		break;
		
	case SMEM_CUR:
		//電流
		sprintf(aTxtBuf,"%03d",inValue);//0x00は無効
		buffer_sharedStatusMem[12] = aTxtBuf[0];
		buffer_sharedStatusMem[13] = aTxtBuf[1];
		break;
		
	case SMEM_SEQ:
		sprintf(aTxtBuf,"%02d",inValue);//0x00は無効
		buffer_sharedStatusMem[17] = aTxtBuf[0];
		buffer_sharedStatusMem[18] = aTxtBuf[1];
		break;
		
	}
}

boolean SDIO_WriteSharedMem_CVReply(word inCVNo, uint8_t inValue)
{
	
	if( gFlag_FlashAir == 0)
	{
		return false;
	}
	
	char aBufCVReply[16];
	
	//@CV,9999,255, (13文字)
	sprintf(aBufCVReply, "@CV,%04d,%03d,", inCVNo, inValue);
	
	for(int i = 0; i < 13; i++)
	{
		buffer_sharedStatusMem[32 + i] = aBufCVReply[i];
	}	
	
	return true;
}

boolean SDIO_WriteSharedMem_Acc()
{
	
	if( gFlag_FlashAir == 0)
	{
		return false;
	}
	
	char aBuf2[2];
	
	for(int i = 0; i < SHAREMEM_ACCSIZE; i++)
	{
		sprintf(aBuf2, "%02x", buffer_accessories[i]);
		
		buffer_sharedStatusMem[48 + i * 2] = aBuf2[1];
		buffer_sharedStatusMem[48 + i * 2 + 1] = aBuf2[0];
	}
	
	return true;
}

void LOCMNG_UpdateAcc(word inNo)
{
	char aBuf2[2];
	
	sprintf(aBuf2, "%02x", buffer_accessories[inNo]);
	
	buffer_sharedStatusMem[48 + inNo * 2] = aBuf2[1];
	buffer_sharedStatusMem[48 + inNo * 2 + 1] = aBuf2[0];
	
}

//FFFF,1,FFFFFFFF



void CalcCurrentOffset()
{
	unsigned short aCurrentSum = 0;
	
	if( gOffset_Current == 0)
	{
		/* 直流中間に電圧が印加されてないと、実際の動作時のADCのオフセット誤差が異なる */
		if( analogRead(PIN_EDC) >= THRESHOLD_EDC)
		{
			
			for( int i = 0; i < 4; i++)
			{
				aCurrentSum = aCurrentSum + (analogRead(PIN_CURRENT) >> ADC_SHIFT);
				delay(10);
			}
			
			//平均値を算出
			gOffset_Current = aCurrentSum >> 2;
		}
	}
	
	/* 検出できない場合は512をオフセット値にセット */
	if( gOffset_Current == 0)
	{
		gOffset_Current = 512;
	}
}

void LOCMNG_Clear(void)
{
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		gLocStatus[i].mAddress = 0;
		gLocStatus[i].mSpeed = 0;
		gLocStatus[i].mDir = 0;
		gLocStatus[i].mFuncBuf1 = 0;
		gLocStatus[i].mFuncBuf2 = 0;
	}
}


byte LOCMNG_SetLocoSpeed(word inAddr, word inSpd)
{
	byte aIndex = 255;
	word aSpeed = 0;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		aSpeed = inSpd >> 2;
		
		if( aSpeed > 255)
		{
			aSpeed = 255;
		}
		
		char aStrAddr[5];
		char aStrSpd[3];
		
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);
		sprintf(aStrSpd, "%02X", aSpeed);
		
		/* FlashAir共有メモリ書き換え */
		buffer_sharedStatusMem[113 + aIndex * 19	] = aStrAddr[0];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 1] = aStrAddr[1];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 2] = aStrAddr[2];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 3] = aStrAddr[3];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 5] = aStrSpd[0];//Spd
		buffer_sharedStatusMem[113 + aIndex * 19 + 6] = aStrSpd[1];//Spd
		
	}
}

byte LOCMNG_SetLocoDir(word inAddr, byte inDir)
{
	byte aIndex = 255;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		
		if( gLocStatus[aIndex].mDir != inDir)
		{
			gLocStatus[aIndex].mDir = inDir;
			
			buffer_sharedStatusMem[113 + aIndex * 19 + 5] = 0x30;//Spd
			buffer_sharedStatusMem[113 + aIndex * 19 + 6] = 0x30;//Spd
		}
		
		char aStrAddr[5];
		char aStrDir;
		
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);
		aStrDir = (gLocStatus[aIndex].mDir == 0) ? '0' : '1';
		
		/* FlashAir共有メモリ書き換え */
		buffer_sharedStatusMem[113 + aIndex * 19	] = aStrAddr[0];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 1] = aStrAddr[1];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 2] = aStrAddr[2];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 3] = aStrAddr[3];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 8] = aStrDir;//Dir
	}
}

byte LOCMNG_SetLocoFunc(word inAddr, byte inNo, byte inPower)
{
	char aStrAddr[5];
	char aStrFunc1[5];
	char aStrFunc2[5];
	byte aIndex = 255;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);

		buffer_sharedStatusMem[113 + aIndex * 19	] = aStrAddr[0];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 1] = aStrAddr[1];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 2] = aStrAddr[2];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 3] = aStrAddr[3];//LocAddr

		
		if( inPower == 0)
		{
			if( inNo < 16)
			{
				gLocStatus[aIndex].mFuncBuf1 = gLocStatus[aIndex].mFuncBuf1 & ~(1 << inNo);
			}
			else
			{
				gLocStatus[aIndex].mFuncBuf2 = gLocStatus[aIndex].mFuncBuf2 & ~(1 << (inNo - 16));
			}
		}
		else
		{
			if( inNo < 16)
			{
				gLocStatus[aIndex].mFuncBuf1 = gLocStatus[aIndex].mFuncBuf1 | (1 << inNo);
			}
			else
			{
				gLocStatus[aIndex].mFuncBuf2 = gLocStatus[aIndex].mFuncBuf2 | (1 << (inNo - 16));
			}
		}
		
		
		sprintf(aStrFunc1, "%04X", gLocStatus[aIndex].mFuncBuf1);
		sprintf(aStrFunc2, "%04X", gLocStatus[aIndex].mFuncBuf2);
		
		
		/* FlashAir共有メモリ書き換え */
		
		
		buffer_sharedStatusMem[113 + aIndex * 19 + 10] = aStrFunc2[0];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 11] = aStrFunc2[1];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 12] = aStrFunc2[2];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 13] = aStrFunc2[3];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 14] = aStrFunc1[0];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 15] = aStrFunc1[1];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 16] = aStrFunc1[2];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 17] = aStrFunc1[3];//Fnc
	}
}
